"""Text exceptions raised by the Forecast.Solar API client."""

import pytest
from aresponses import ResponsesMockServer

from forecast_solar import (
    ForecastSolar,
    ForecastSolarAuthenticationError,
    ForecastSolarConfigError,
    ForecastSolarConnectionError,
    ForecastSolarRatelimitError,
    ForecastSolarRequestError,
)

from . import load_fixtures


async def test_status_400(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test response status 400."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=400,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    with pytest.raises(ForecastSolarRequestError):
        assert await forecast_client._request("test")


async def test_status_401(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test response status 401 or 403."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=401,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    with pytest.raises(ForecastSolarAuthenticationError):
        assert await forecast_client._request("test")


async def test_status_422(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test response status 422."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=422,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    with pytest.raises(ForecastSolarConfigError):
        assert await forecast_client._request("test")


async def test_status_429(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test response status 429."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=429,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("ratelimit.json"),
        ),
    )
    with pytest.raises(ForecastSolarRatelimitError):
        assert await forecast_client._request("test")


async def test_status_502(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test response status 502 or 503."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=502,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    with pytest.raises(ForecastSolarConnectionError):
        assert await forecast_client._request("test")



"""Fixtures for the Forecast.Solar tests."""

from collections.abc import AsyncGenerator

import pytest
from aiohttp import ClientSession

from forecast_solar import ForecastSolar


@pytest.fixture(name="forecast_client")
async def client() -> AsyncGenerator[ForecastSolar, None]:
    """Return a Forecast.Solar client."""
    async with (
        ClientSession() as session,
        ForecastSolar(
            latitude=52.16,
            longitude=4.47,
            declination=20,
            azimuth=10,
            kwp=2.160,
            damping=0,
            horizon="0,0,0,10,10,20,20,30,30",
            session=session,
        ) as forecast_client,
    ):
        yield forecast_client


@pytest.fixture(name="forecast_key_client")
async def client_api_key() -> AsyncGenerator[ForecastSolar, None]:
    """Return a Forecast.Solar client."""
    async with (
        ClientSession() as session,
        ForecastSolar(
            api_key="myapikey",
            latitude=52.16,
            longitude=4.47,
            declination=20,
            azimuth=10,
            kwp=2.160,
            damping_morning=0,
            damping_evening=0,
            inverter=1.300,
            session=session,
        ) as forecast_key_client,
    ):
        yield forecast_key_client



"""Tests for Forecast.Solar."""

# pylint: disable=protected-access

import pytest
from aresponses import ResponsesMockServer

from forecast_solar import (
    ForecastSolar,
    ForecastSolarError,
)

from . import load_fixtures


async def test_json_request(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test JSON response is handled correctly."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    response = await forecast_client._request("test")
    assert response is not None
    await forecast_client.close()


async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test internal session is handled correctly."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    async with ForecastSolar(
        latitude=52.16,
        longitude=4.47,
        declination=20,
        azimuth=10,
        kwp=2.160,
        damping=0,
        horizon="0,0,0,10,10,20,20,30,30",
    ) as client:
        await client._request("test")


async def test_content_type(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test content type error handling."""
    aresponses.add(
        "api.forecast.solar",
        "/test",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "blabla/blabla",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
        ),
    )
    with pytest.raises(ForecastSolarError):
        assert await forecast_client._request("test")



"""Asynchronous Python client for the Forecast.Solar API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Self

from aiohttp import ClientSession
from yarl import URL

from .exceptions import (
    ForecastSolarAuthenticationError,
    ForecastSolarConfigError,
    ForecastSolarConnectionError,
    ForecastSolarError,
    ForecastSolarRatelimitError,
    ForecastSolarRequestError,
)
from .models import Estimate, Ratelimit


@dataclass
class ForecastSolar:
    """Main class for handling connections with the Forecast.Solar API."""

    azimuth: float
    declination: float
    kwp: float
    latitude: float
    longitude: float

    api_key: str | None = None
    damping: float = 0
    damping_morning: float | None = None
    damping_evening: float | None = None
    horizon: str | None = None

    session: ClientSession | None = None
    ratelimit: Ratelimit | None = None
    inverter: float | None = None
    _close_session: bool = False
    _base_url = URL("https://api.forecast.solar")

    async def _request(
        self,
        uri: str,
        *,
        rate_limit: bool = True,
        authenticate: bool = True,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the Forecast.Solar API.

        A generic method for sending/handling HTTP requests done against
        the Forecast.Solar API.

        Args:
        ----
            uri: Request URI, for example, 'estimate'
            rate_limit: Parse rate limit from response. Set to False for
                endpoints that are missing rate limiting headers in response.
            authenticate: Prefix request with api_key. Set to False for
                endpoints that do not provide authentication.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from
            the Forecast.Solar API.

        Raises:
        ------
            ForecastSolarAuthenticationError: If the API key is invalid.
            ForecastSolarConnectionError: An error occurred while communicating
                with the Forecast.Solar API.
            ForecastSolarError: Received an unexpected response from the
                Forecast.Solar API.
            ForecastSolarRequestError: There is something wrong with the
                variables used in the request.
            ForecastSolarRatelimitError: The number of requests has exceeded
                the rate limit of the Forecast.Solar API.

        """
        # Add API key if one is provided
        if authenticate and self.api_key is not None:
            url = self._base_url.with_path(f"{self.api_key}/")
        else:
            url = self._base_url

        url = url.join(URL(uri))

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        response = await self.session.request(
            "GET",
            url,
            params=params,
            ssl=False,
        )

        if response.status in (502, 503):
            raise ForecastSolarConnectionError("The Forecast.Solar API is unreachable")

        if response.status == 400:
            data = await response.json()
            raise ForecastSolarRequestError(data["message"])

        if response.status in (401, 403):
            data = await response.json()
            raise ForecastSolarAuthenticationError(data["message"])

        if response.status == 422:
            data = await response.json()
            raise ForecastSolarConfigError(data["message"])

        if response.status == 429:
            data = await response.json()
            raise ForecastSolarRatelimitError(data["message"])

        if rate_limit and response.status == 200:
            self.ratelimit = Ratelimit.from_response(response)

        response.raise_for_status()

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            raise ForecastSolarError(
                "Unexpected response from the Forecast.Solar API",
                {"Content-Type": content_type, "response": text},
            )

        return await response.json()

    async def validate_plane(self) -> bool:
        """Validate plane by calling the Forecast.Solar API.

        Returns
        -------
            True, if plane is valid.

        """
        await self._request(
            f"check/{self.latitude}/{self.longitude}"
            f"/{self.declination}/{self.azimuth}/{self.kwp}",
            rate_limit=False,
            authenticate=False,
        )

        return True

    async def validate_api_key(self) -> bool:
        """Validate api key by calling the Forecast.Solar API.

        Returns
        -------
            True, if api key is valid

        """
        await self._request("info", rate_limit=False)

        return True

    async def estimate(self, actual: float = 0) -> Estimate:
        """Get solar production estimations from the Forecast.Solar API.

        Args:
        ----
            actual: The production for the day in kWh so far. Used to improve
                the estimation for the current day if an API key is provided.

        Returns:
        -------
            A Estimate object, with a estimated production forecast.

        """
        params = {"time": "utc", "damping": str(self.damping)}
        if self.inverter is not None:
            params["inverter"] = str(self.inverter)
        if self.horizon is not None:
            params["horizon"] = str(self.horizon)
        if self.damping_morning is not None and self.damping_evening is not None:
            params["damping_morning"] = str(self.damping_morning)
            params["damping_evening"] = str(self.damping_evening)
        if self.api_key is not None:
            params["actual"] = str(actual)

        data = await self._request(
            f"estimate/{self.latitude}/{self.longitude}"
            f"/{self.declination}/{self.azimuth}/{self.kwp}",
            params=params,
        )

        return Estimate.from_dict(data)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The ForecastSolar object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Example of how to get an estimate from the Forecast.Solar API."""

import asyncio
from datetime import UTC, datetime, timedelta
from pprint import pprint  # noqa: F401

from forecast_solar import ForecastSolar, ForecastSolarRatelimitError


async def main() -> None:
    """Get an estimate from the Forecast.Solar API."""
    async with ForecastSolar(
        latitude=52.16,
        longitude=4.47,
        declination=20,
        azimuth=10,
        kwp=2.160,
        damping=0,
        horizon="0,0,0,10,10,20,20,30,30",
    ) as forecast:
        try:
            estimate = await forecast.estimate()
        except ForecastSolarRatelimitError as err:
            print("Ratelimit reached")
            print(f"Rate limit resets at {err.reset_at}")
            reset_period = err.reset_at - datetime.now(UTC)
            # Strip microseconds as they are not informative
            reset_period -= timedelta(microseconds=reset_period.microseconds)
            print(f"That's in {reset_period}")
            return

        # Uncomment this if you want to see what's in the estimate arrays
        # pprint(dataclasses.asdict(estimate))
        print()
        print(f"energy_production_today: {estimate.energy_production_today}")
        print(
            "energy_production_today_remaining: "
            f"{estimate.energy_production_today_remaining}"
        )
        print(
            f"power_highest_peak_time_today: {estimate.power_highest_peak_time_today}"
        )
        print(f"energy_production_tomorrow: {estimate.energy_production_tomorrow}")
        print(
            "power_highest_peak_time_tomorrow: "
            f"{estimate.power_highest_peak_time_tomorrow}"
        )
        print()
        print(f"power_production_now: {estimate.power_production_now}")
        print(
            "power_production in 1 hour: "
            f"{estimate.power_production_at_time(estimate.now() + timedelta(hours=1))}"
        )
        print(
            "power_production in 6 hours: "
            f"{estimate.power_production_at_time(estimate.now() + timedelta(hours=6))}"
        )
        print(
            "power_production in 12 hours: "
            f"{estimate.power_production_at_time(estimate.now() + timedelta(hours=12))}"
        )
        print(
            "power_production in 24 hours: "
            f"{estimate.power_production_at_time(estimate.now() + timedelta(hours=24))}"
        )
        print()
        print(f"energy_current_hour: {estimate.energy_current_hour}")
        print(f"energy_production next hour: {estimate.sum_energy_production(1)}")
        print(f"energy_production next 6 hours: {estimate.sum_energy_production(6)}")
        print(f"energy_production next 12 hours: {estimate.sum_energy_production(12)}")
        print(f"energy_production next 24 hours: {estimate.sum_energy_production(24)}")
        print(f"timezone: {estimate.timezone}")
        print(f"account_type: {estimate.account_type}")
        print(forecast.ratelimit)


if __name__ == "__main__":
    asyncio.run(main())



"""Test the models."""

from datetime import datetime

import pytest
from aresponses import ResponsesMockServer
from syrupy.assertion import SnapshotAssertion

from forecast_solar import AccountType, Estimate, ForecastSolar

from . import load_fixtures


@pytest.mark.freeze_time("2024-04-26T12:00:00+02:00")
async def test_estimated_forecast(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    forecast_client: ForecastSolar,
) -> None:
    """Test estimated forecast."""
    aresponses.add(
        "api.forecast.solar",
        "/estimate/52.16/4.47/20/10/2.16",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("forecast.json"),
        ),
    )
    forecast: Estimate = await forecast_client.estimate()
    assert forecast == snapshot
    assert forecast.timezone == "Europe/Amsterdam"
    assert forecast.account_type == AccountType.PUBLIC

    assert forecast.energy_production_today == 6660
    assert forecast.energy_production_tomorrow == 5338

    assert forecast.power_production_now == 773
    assert forecast.energy_production_today_remaining == 4144
    assert forecast.energy_current_hour == 821

    assert forecast.power_highest_peak_time_today == datetime.fromisoformat(
        "2024-04-26T11:00:00+02:00"
    )
    assert forecast.power_highest_peak_time_tomorrow == datetime.fromisoformat(
        "2024-04-27T12:00:00+02:00"
    )

    assert forecast.sum_energy_production(1) == 742
    assert forecast.sum_energy_production(6) == 3093
    assert forecast.sum_energy_production(12) == 3323
    assert forecast.sum_energy_production(24) == 5633


@pytest.mark.freeze_time("2024-04-27T07:00:00+02:00")
async def test_estimated_forecast_with_subscription(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    forecast_key_client: ForecastSolar,
) -> None:
    """Test estimated forecast."""
    aresponses.add(
        "api.forecast.solar",
        "/myapikey/estimate/52.16/4.47/20/10/2.16",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "60",
                "X-Ratelimit-Period": "3600",
            },
            text=load_fixtures("forecast_personal.json"),
        ),
    )
    forecast: Estimate = await forecast_key_client.estimate()
    assert forecast == snapshot
    assert forecast.timezone == "Europe/Amsterdam"
    assert forecast.account_type == AccountType.PERSONAL

    assert forecast.energy_production_today == 5788
    assert forecast.energy_production_tomorrow == 7507

    assert forecast.power_production_now == 92
    assert forecast.energy_production_today_remaining == 5783
    assert forecast.energy_current_hour == 96

    assert forecast.power_highest_peak_time_today == datetime.fromisoformat(
        "2024-04-27T13:30:00+02:00"
    )
    assert forecast.power_highest_peak_time_tomorrow == datetime.fromisoformat(
        "2024-04-28T14:30:00+02:00"
    )

    assert forecast.sum_energy_production(1) == 216
    assert forecast.sum_energy_production(6) == 2802
    assert forecast.sum_energy_production(12) == 5582
    assert forecast.sum_energy_production(24) == 5784


@pytest.mark.freeze_time("2024-04-27T07:00:00+02:00")
async def test_estimated_forecast_with_subscription_and_actual_value(
    aresponses: ResponsesMockServer,
    snapshot: SnapshotAssertion,
    forecast_key_client: ForecastSolar,
) -> None:
    """Test estimated forecast."""
    aresponses.add(
        "api.forecast.solar",
        "/myapikey/estimate/52.16/4.47/20/10/2.16",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "60",
                "X-Ratelimit-Period": "3600",
            },
            text=load_fixtures("forecast_personal.json"),
        ),
    )
    forecast: Estimate = await forecast_key_client.estimate(actual=2.300)
    assert forecast == snapshot
    assert forecast.timezone == "Europe/Amsterdam"
    assert forecast.account_type == AccountType.PERSONAL

    assert forecast.energy_production_today == 5788
    assert forecast.energy_production_tomorrow == 7507

    assert forecast.sum_energy_production(1) == 216
    assert forecast.sum_energy_production(6) == 2802
    assert forecast.sum_energy_production(12) == 5582
    assert forecast.sum_energy_production(24) == 5784


async def test_api_key_validation(
    aresponses: ResponsesMockServer,
    forecast_key_client: ForecastSolar,
) -> None:
    """Test API key validation."""
    aresponses.add(
        "api.forecast.solar",
        "/myapikey/info",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("validate_key.json"),
        ),
    )
    assert await forecast_key_client.validate_api_key() is True


async def test_plane_validation(
    aresponses: ResponsesMockServer,
    forecast_client: ForecastSolar,
) -> None:
    """Test plane validation."""
    aresponses.add(
        "api.forecast.solar",
        "/check/52.16/4.47/20/10/2.16",
        "GET",
        aresponses.Response(
            status=200,
            headers={
                "Content-Type": "application/json",
                "X-Ratelimit-Limit": "10",
                "X-Ratelimit-Period": "1",
            },
            text=load_fixtures("validate_plane.json"),
        ),
    )
    assert await forecast_client.validate_plane() is True
